@@ -94,6 +94,7 @@ bsp=function(x, xk, qw, n=3L) {
 #' @param tol Numerical scalar, relative tolerance for small singular values
 #'   that should be considered as 0 if \code{s[i] <= tol*s[1]}. This parameter
 #'   is ignored if estSD=FALSE (1.e-10 by default).
+#' @param regular_grid Logical scalar, should regular knot grid be used when xki=NULL (FALSE by default)
 #' @return Function, smoothing B-splines
 #'   respecting optional constraints (generated by \code{par2bsp()}).
 #' @details
@@ -129,27 +130,28 @@ bsp=function(x, xk, qw, n=3L) {
 #' @importFrom stats sd
 #' @aliases fitsmbsp smbsp
 #' @export
-smbsp=function(x, y, n=3L, xki=NULL, nki=1L, lieq=NULL, monotone=0, positive=0, mat=NULL, estSD=FALSE, tol=1.e-10) {
+smbsp=function(x, y, n=3L, xki=NULL, nki=1L, lieq=NULL, monotone=0, positive=0, mat=NULL, estSD=FALSE, tol=1.e-10, regular_grid=FALSE) {
     y=as.matrix(y)
     nx=length(x)
     nc=ncol(y)
     stopifnot(nx == nrow(y))
-
+    rax=range(x)
     if (is.null(xki)) {
         # build default xki
-        #dxk=diff(range(x))/(nki+1)
-        #xki=x[1L]+seq_len(nki)*dxk
-        xki=iknots(x, y, nki=nki, n=n)
+        if (regular_grid) {
+            xki=seq(x[1L], x[nx], length.out=nki+2L)[-c(1L, nki+2L)]
+        } else {
+            xki=iknots(x, y, nki=nki, n=n, lenfit=0L, smooth=TRUE)
+        }
     } else {
+        # remove end- and outside- points
+        if (any(iout <- xki <= rax[1L] | xki >= rax[2L])) {
+            warning("smbsp: following xki values are removed as laying outside or on the ends of x interval: ", paste0(xki[iout], collapse="; "))
+            xki=xki[!iout]
+        }
         nki=length(xki)
     }
-    #stopifnot(nki > 0)
-    if (nki > 0L) {
-        stopifnot("xki is not strictly inside of x range"=all((range(xki)-range(x))*c(1,-1) > 0))
-        xk=c(rep(x[1L], n+1L), xki, rep(x[nx], n+1L))
-    } else {
-        xk=c(rep(x[1L], n+1L), rep(x[nx], n+1L))
-    }
+    xk=c(rep(x[1L], n+1L), xki, rep(x[nx], n+1L))
     
     nk=length(xk)
     nw=nk-n-1L
@@ -159,10 +161,17 @@ smbsp=function(x, y, n=3L, xki=NULL, nki=1L, lieq=NULL, monotone=0, positive=0,
     if (!is.null(lieq)) {
         stopifnot(length(lieq) == ncol(y))
         # todo: bezn() only on unique x from lieq
-        liece=lapply(lieq, function(meq) {
+        if (length(nm_lieq <- names(lieq)) == 0L)
+            nm_lieq=paste0("eq_", seq_along(lieq))
+        
+        liece=mapply(nm_lieq, lieq, SIMPLIFY=FALSE, FUN=function(nm, meq) {
             if (is.null(meq) || length(meq) == 0L) {
                 NULL
             } else {
+                if (any(ibad <- meq[,1L] < rax[1L] | meq[,1L] > rax[2L]))
+                    stop("The following arguments for equality constraint '", nm,
+                        "' are out of x range [", rax[1L], "; ", rax[2L], "]:\n\t",
+                        paste0("x=", meq[ibad,1L], collapse="\n\t"), call.=FALSE)
                 list(e=bsc(meq[,1L], n=n, xk), ce=meq[,2L])
             }
         })
@@ -237,7 +246,9 @@ smbsp=function(x, y, n=3L, xki=NULL, nki=1L, lieq=NULL, monotone=0, positive=0,
         qw=structure(vapply(seq_len(nc), function(ic) {
             m=monotone[ic]
             p=positive[ic]
-            nlsic::lsie_ln(qra, y[,ic], u=uco[[m]][[p]]$u, co=uco[[m]][[p]]$co, e=liece[[ic]]$e, ce=liece[[ic]]$ce)
+            e=if (!is.null(liece)) liece[[ic]]$e else NULL
+            ce=if (!is.null(liece)) liece[[ic]]$ce else NULL
+            nlsic::lsie_ln(qra, y[,ic], u=uco[[m]][[p]]$u, co=uco[[m]][[p]]$co, e=e, ce=ce)
         }, double(nw)), dim=c(nw, nc), dimnames=list(NULL, colnames(y)))
         #browser()
     } else {
@@ -296,22 +307,41 @@ smbsp=function(x, y, n=3L, xki=NULL, nki=1L, lieq=NULL, monotone=0, positive=0,
 #'  lines(x, f(x))
 #' @importFrom nlsic nlsic lsi_ln
 #' @export
-fitsmbsp=function(x, y, n=3L, xki=NULL, nki=1L, lieq=NULL, monotone=0, positive=0, control=list(), estSD=FALSE, tol=1.e-10) {
+fitsmbsp=function(x, y, n=3L, xki=NULL, nki=1L, lieq=NULL, monotone=0, positive=0, control=list(), estSD=FALSE, tol=1.e-10, regular_grid=FALSE) {
     np=length(x)
     y=as.matrix(y)
     stopifnot(nrow(y) == np)
+    rax=range(x)
     if (is.null(xki)) {
-        xki=iknots(x, y, n=n, nki=nki)
+        # build default xki
+        if (regular_grid) {
+            # no knot optimization
+            return(smbsp(x, y, n=n, xki=NULL, nki=nki, lieq, monotone, positive, estSD=estSD, tol=tol, regular_grid=regular_grid))
+        } else {
+            if (np > n+1L) {
+                xki=iknots(x, y, nki=nki, n=n, lenfit=0L, smooth=TRUE)
+            } else {
+                # start with regular grid
+                xki=seq(x[1L], x[np], length.out=nki+2L)[-c(1L, nki+2L)]
+            }
+        }
+    } else {
+        # remove end- and outside- points
+        if (any(iout <- xki <= rax[1L] | xki >= rax[2L])) {
+            warning("smbsp: following xki values are removed as laying outside or on the ends of x interval: ", paste0(xki[iout], collapse="; "))
+            xki=xki[!iout]
+        }
+        nki=length(xki)
     }
     nki=length(xki)
     if (nki == 0L)
-        return(smbsp(x, y, n=n, xki=NULL, nki=0L, lieq, monotone, positive))
+        return(smbsp(x, y, n=n, xki=NULL, nki=0L, lieq, monotone, positive, estSD=estSD, tol=tol))
     # inequalities
     # p[1] >= tail(x_1, 1)+epsx; p[i]+epsx <= p[i+1]; p[nki]+epsx <= x_n
     dx=diff(x)
     dx=dx[dx != 0.]
     stopifnot(dx > 0.)
-    epsx=min(dx)*0.1
+    epsx=min(dx)
     u=matrix(0., nrow=nki+1L, ncol=nki)
     co=double(nrow(u))
     u[1L,1L]=1.; co[1L]=x[1L]+epsx
@@ -324,17 +354,17 @@ fitsmbsp=function(x, y, n=3L, xki=NULL, nki=1L, lieq=NULL, monotone=0, positive=
     fresid=function(xki, cjac) {
         if (cjac) {
             li=bsc(x, c(x1, xki, x2), n=n, cjac=TRUE)
-            f=smbsp(x, y, n=n, xki=xki, nki=0, lieq=lieq, monotone=monotone, positive=positive, mat=li$mat)
+            f=smbsp(x, y, n=n, xki=xki, nki=0, lieq=lieq, monotone=monotone, positive=positive, mat=li$mat, tol=tol)
             res=f(x)-y
             jac=jacw(li$jac, bsppar(f)$qw)[,,(n+1L)+seq_len(nki), drop=FALSE]
             dim(jac)=c(np*ncol(y), nki)
             list(res=res, jacobian=jac)
         } else {
-            list(res=smbsp(x, y, n=n, xki=xki, nki=0L, lieq=lieq, monotone=monotone, positive=positive)(x)-y)
+            list(res=smbsp(x, y, n=n, xki=xki, nki=0L, lieq=lieq, monotone=monotone, positive=positive, tol=tol)(x)-y)
         }
     }
     if (is.null(control$errx))
-        control$errx=epsx*0.1
+        control$errx=epsx*0.01
     fit=nlsic::nlsic(xki, fresid,
         u=u, co=co, control=control, flsi=nlsic::lsi_ln)
     if (fit$error != 0)
@@ -554,12 +584,11 @@ bsppar=function(f) {
 #' @param sdqw Numeric Matrix, SD of qw thus having the same dimension
 #'   as qw (can be estimated in \code{\link{smbsp}}).
 #' @return Function, calculating B-splines at arbitrary points and having
-#'  interface \code{f(x, select)} where \code{x} is a vector of abscissa points.
+#'  interface \code{f(x, select, fsd=0.)} where \code{x} is a vector of abscissa points.
 #'  Parameter \code{select} is passed to
-#'  \code{qw[, select, drop=FALSE]} and can be missing. This function will return
-#'  a matrix of size \code{length(x) x ncol(qw)} if \code{select} is missing. Elsewhere,
-#'  a number of column will depend on \code{select} parameter. Column names in
-#'  the result matrix will be inherited from \code{qw}.
+#'  \code{qw[, select, drop=FALSE]} and can be missing or NULL in which case all columns of qw are used.
+#'  \code{fsd} is a factor for SD to be added to the B-spline. Column names in
+#'  the result matrix returned by \code{f()} will be inherited from \code{qw}.
 #' @export
 par2bsp=function(n, qw, xk, covqw=NULL, sdy=NULL, sdqw=NULL)
     local({
@@ -572,12 +601,12 @@ par2bsp=function(n, qw, xk, covqw=NULL, sdy=NULL, sdqw=NULL)
         sdqw=sdqw
         function(x, select, fsd=0.) {
             if (fsd == 0.) {
-                if (base::missing(select)) bspline::bsp(x, xk, qw, n=n) else
+                if (base::missing(select) || is.null(select)) bspline::bsp(x, xk, qw, n=n) else
                 bspline::bsp(x, xk, qw[, select, drop=FALSE], n=n)
             } else {
                 if (is.null(sdqw))
                     stop("B-spline is asked for fsd != 0 but sdqw is NULL")
-                if (base::missing(select)) bspline::bsp(x, xk, qw+fsd*sdqw, n=n) else
+                if (base::missing(select) || is.null(select)) bspline::bsp(x, xk, qw+fsd*sdqw, n=n) else
                 bspline::bsp(x, xk, qw[, select, drop=FALSE]+fsd*sdqw[, select, drop=FALSE], n=n)
             }
         }
@@ -605,28 +634,51 @@ diffn=function(m, ndiff=1L) {
     colnames(res)=colnames(m)
     res
 }
+#' Smoothing vector or columns of matrix y on possibly irregular grid x by Laplacian.
+#' The end ponits remain untouched.
+#' @noRd
+#' @keywords internal
+
+smooth_lap=function(x, y) {
+    ymat=is.matrix(y)
+    y=as.matrix(y)
+    stopifnot(length(x) == nrow(y))
+    nx=length(x)
+    dx=diff(x)
+    ym=y[-1L,,drop=FALSE]+y[-nx,,drop=FALSE]
+    y[-c(1L, nx),]=0.5*(ym[-nx+1L,]*dx[-1L] + ym[-1L,]*dx[-nx+1L])/(dx[-1L]+dx[-nx+1L])
+    if (ymat) y else y[,1L]
+}
+
 #' Estimate internal knot positions equalizing jumps in n-th derivative
 #'
-#' Normalized total variation of n-th finite differences is calculated for each column in
-#' \code{y} then averaged. These averaged values are fitted by a linear spline to
-#' find knot positions that equalize the jumps of n-th derivative.\cr
-#' NB. This function is used internally in \code{(fit)smbsp()} and a priori
-#' has no interest to be called directly by user.
+#' Normalized total variation of n-th finite
+#' differences is calculated for each column in
+#' \code{y} then averaged. These averaged values are smoothed (if requested)
+#' and sampled to find knot positions which equalize the jumps of n-th
+#' derivative. If any of found knots fall inside a same x interval, TV is
+#' smoothed till no such situation occurs.
+#'
 #' @param x Numeric vector
 #' @param y Numeric vector or matrix
 #' @param nki Integer scalar, number of internal knots to estimate (1 by default)
 #' @param n Integer scalar, polynomial order of B-spline (3 by default)
-#' @param lenfit Integer scalar, length of knots for linear spline to fit the total variation)
+#' @param lenfit Integer scalar, length of knots for linear spline to fit the total variation if greater than 0, otherwise full linear interpolation is used (12 by default)
+#' @param smooth Logical scalar, should total variation be smoothed? (FALSE by default)
 #' @return Numeric vector, estimated knot positions
 #' @export
-iknots=function(x, y, nki=1L, n=3L, lenfit=12L) {
+iknots=function(x, y, nki=1L, n=3L, lenfit=12L, smooth=FALSE) {
     stopifnot(nki >= 0L)
+    stopifnot(nki <= length(x)-2L)
     if (nki == 0L)
         return(double(0L))
+    if (nki == length(x)-2L)
+        return(x[-c(1L, length(x))])
     y=as.matrix(y)
     xy=cbind(x, y)
     dubx=duplicated(x)
     if (any(dubx)) {
+        # make x unique by averaging y on the repeated points
         ix=split(seq_along(x), x)
         xy=t(vapply(ix, function(ii) if (length(ii) > 1L) colMeans(xy[ii,,drop=FALSE]) else xy[ii,], xy[1L,]))
     }
@@ -634,32 +686,56 @@ iknots=function(x, y, nki=1L, n=3L, lenfit=12L) {
     dxy=diffn(dxyn, 1L)
     dy=dxy[,-1L, drop=FALSE]
     xtv=dxyn[,1L]
+    # linearly extend xtv to cover the same interval as x
+    ra=range(xtv)
+    rax=range(x)
+    a=diff(rax)/diff(ra)
+    b=x[1L]-a*xtv[1L]
+    xtve=a*xtv+b
     tv=rbind(0., arrApply::arrApply(abs(dy)*diff(xtv), 1L, "cumsum"))
     tv=arrApply::arrApply(tv, 2L, "multv", v=1./tv[nrow(tv),])
     tv[!is.finite(tv)]=0.
     tv=rowSums(tv)
     tv=tv/tv[length(tv)]
-    ra=range(xtv)
-    ftv=smbsp(xtv, tv, xki=xtv[1L]+diff(ra)*seq(0., 1., length.out=lenfit)[2L:(lenfit-1L)], n=1L, lieq=list(cbind(ra,0:1)), monotone=1)
-    par=bsppar(ftv)
-    if (FALSE) {
-        print(c("knots qw=", par$qw))
-        print(c("knots dqw=", diff(par$qw)))
+    ntv=length(tv)
+    
+    if (smooth)
+        tv=smooth_lap(xtve, tv)
+
+    etv=seq(0., 1., length.out=nki+2L) # equalized tv; +2 for excluding end points
+    knots_too_close=TRUE
+    while (knots_too_close) {
+        if (lenfit > 0L) {
+            ftv=smbsp(xtve, tv, xki=seq(xtve[1L], xtve[ntv], length.out=lenfit+2L)[-c(1L, lenfit+2L)], n=1L, lieq=list(cbind(rax,0:1)), monotone=1)
+            par=bsppar(ftv)
+            if (FALSE) {
+                print(c("knots qw=", par$qw))
+                print(c("knots dqw=", diff(par$qw)))
+            }
+            qw=par$qw
+            xk=par$xk
+        } else {
+            # full linear interpolation
+            qw=tv
+            xk=xtve
+        }
+        dq=diff(qw)
+        dq[dq < 0.]=0. # to force monotonicity despite round off errors
+        qw=c(0., cumsum(dq))
+        qw=qw/qw[length(qw)] # to make qw end up in 1
+        ik=findInterval(etv, qw, rightmost.closed=TRUE)
+        i=seq_along(etv)
+        k=ik[i]
+        x1=xk[k]
+        x2=xk[k+1L]
+        y1=qw[k]
+        y2=qw[k+1L]
+        xki=x1+(x2-x1)*(etv-y1)/(y2-y1)
+        knots_too_close=anyDuplicated(findInterval(xki, x, rightmost.closed=TRUE)) > 0L
+        if (knots_too_close)
+            tv=smooth_lap(xtve, tv)
     }
-    etv=seq(0., 1., length.out=nki+2L)[c(-1L, -(nki+2L))] # equalized tv
-    qw=par$qw
-    dq=diff(qw)
-    dq[dq < 0.]=0. # to force monotonicity despite round off errors
-    qw=c(0., cumsum(dq))
-    qw=qw/qw[length(qw)] # to make qw end up in 1
-    ik=findInterval(etv, qw, rightmost.closed=TRUE)
-    i=seq_along(etv)
-    k=ik[i]
-    x1=par$xk[k+1L]
-    x2=par$xk[k+2L]
-    y1=qw[k]
-    y2=qw[k+1L]
-    x1+(x2-x1)*(etv-y1)/(y2-y1)
+    xki[-c(1L, nki+2L)]
 }
 
 #' nD B-curve governed by (x,y,...) control points.
